#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#    Derived from:
#    https://opendev.org/openstack/etcd3gw/src/tag/2.1.0/etcd3gw/utils.py

import base64

__all__ = [
    "DEFAULT_TIMEOUT",
    "LOCK_PREFIX",
    "_encode",
    "_decode",
    "_increment_last_byte",
]

DEFAULT_TIMEOUT = 30
LOCK_PREFIX = "/locks/"
bytes_types = (bytes, bytearray)


def _encode(data):
    """Encode the given data using base-64.

    This function takes a data as an argument and encodes it using base-64.
    It returns a base-64 encoded string. If the data is not a bytes object,
    it converts it to a bytes object using latin-1 encoding.

    Example:
        >>> _encode("hello")
        'aGVsbG8='

    Args:
        data (bytes or str): The data to encode.

    Returns:
        str: The base-64 encoded string.
    """
    if not isinstance(data, bytes_types):
        data = str(data).encode("latin-1")
    return base64.b64encode(data).decode("utf-8")


def _decode(data):
    """Decode the base-64 encoded string.

    This function takes a data as an argument and decodes it using base-64.
    It returns the decoded data. If the data is not a bytes object,
    it converts it to a bytes object using latin-1 encoding.

    Example:
        >>> _decode('aGVsbG8=')
        b'hello'

    Args:
        data (bytes or str): The base-64 encoded string.

    Returns:
        bytes: The decoded data.
    """
    if not isinstance(data, bytes_types):
        data = str(data).encode("latin-1")
    return base64.b64decode(data.decode("utf-8"))


def _increment_last_byte(data):
    """Get the last byte in the array and increment it.

    This function takes a data as an argument and increments the last byte in the
    array by one. It returns the modified data. If the data is not a bytes object,
    it converts it to a bytes object using utf-8 or latin-1 encoding.

    Example:
        >>> _increment_last_byte(b'foo')
        b'fop'

    Args:
        data (bytes or str): The data to increment.

    Returns:
        bytes: The modified data.
    """
    if not isinstance(data, bytes_types):
        if isinstance(data, str):
            data = data.encode("utf-8")
        else:
            data = str(data).encode("latin-1")
    s = bytearray(data)
    s[-1] = s[-1] + 1
    return bytes(s)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#    Derived from:
#    https://opendev.org/openstack/etcd3gw/src/tag/2.1.0/etcd3gw/__init__.py

from . import utils
from .async_client import AsyncEtcd3Client, async_client
from .async_lease import AsyncLease
from .async_lock import AsyncLock

__version__ = "0.7"

__all__ = (
    "AsyncEtcd3Client",
    "AsyncLease",
    "AsyncLock",
    "async_client",
    "utils",
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import asyncio
import os
import uuid
from time import perf_counter

import pytest

from async_etcd3gw import AsyncEtcd3Client, exceptions, utils
from async_etcd3gw.async_client import DEFAULT_API_PATH

ETCD_HOST = os.environ.get("ETCD_HOST", "localhost")
API_PATH = os.environ.get("API_PATH", DEFAULT_API_PATH)


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_status():
    async with AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH) as client:
        response = await client.status()
        assert response is not None
        assert "version" in response
        assert "header" in response
        assert "cluster_id" in response["header"]


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_members():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    response = await client.members()
    assert len(response) > 0
    assert "clientURLs" in response[0]
    assert "peerURLs" in response[0]
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_with_keys_and_values():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    assert await client.put("foo0", "bar0")
    assert await client.put("foo1", 2001)
    assert await client.put("foo2", b"bar2")

    assert [b"bar0"] == await client.get("foo0")
    assert [b"2001"] == await client.get("foo1")
    assert [b"bar2"] == await client.get("foo2")

    assert await client.delete("foo0")
    assert [] == await client.get("foo0")

    assert not await client.delete("foo0")
    assert len(await client.get_all()) > 0
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_get_and_delete_prefix():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    for i in range(20):
        await client.put(f"/doot1/range{i}", "i am a range")

    values = list(await client.get_prefix("/doot1/range"))
    assert len(values) == 20
    for value, metadata in values:
        assert b"i am a range" == value
        assert metadata["key"].startswith(b"/doot1/range")

    assert await client.delete_prefix("/doot1/range")
    values = list(await client.get_prefix("/doot1/range"))
    assert len(values) == 0
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_get_prefix_sort_order():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)

    def remove_prefix(string, prefix):
        return string[len(prefix) :]

    initial_keys = "abcde"
    initial_values = "qwert"

    for k, v in zip(initial_keys, initial_values):
        await client.put("/doot2/{}".format(k), v)

    keys = b""
    for value, meta in await client.get_prefix("/doot2", sort_order="ascend"):
        keys += remove_prefix(meta["key"], "/doot2/")

    assert keys == initial_keys.encode("latin-1")

    reverse_keys = b""
    for value, meta in await client.get_prefix("/doot2", sort_order="descend"):
        reverse_keys += remove_prefix(meta["key"], "/doot2/")

    assert reverse_keys == "".join(reversed(initial_keys)).encode("latin-1")
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_get_prefix_sort_order_explicit_sort_target_key():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)

    def remove_prefix(string, prefix):
        return string[len(prefix) :]

    initial_keys_ordered = "abcde"
    initial_keys = "aebdc"
    initial_values = "qwert"

    for k, v in zip(initial_keys, initial_values):
        await client.put("/doot2/{}".format(k), v)

    keys = b""
    for value, meta in await client.get_prefix("/doot2", sort_order="ascend", sort_target="key"):
        keys += remove_prefix(meta["key"], "/doot2/")

    assert keys == initial_keys_ordered.encode("latin-1")

    reverse_keys = b""
    for value, meta in await client.get_prefix("/doot2", sort_order="descend", sort_target="key"):
        reverse_keys += remove_prefix(meta["key"], "/doot2/")

    assert reverse_keys == "".join(reversed(initial_keys_ordered)).encode("latin-1")
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_get_prefix_sort_order_explicit_sort_target_rev():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)

    def remove_prefix(string, prefix):
        return string[len(prefix) :]

    initial_keys = "aebdc"
    initial_values = "qwert"

    for k, v in zip(initial_keys, initial_values):
        await client.put("/expsortmod/{}".format(k), v)

    keys = b""
    for value, meta in await client.get_prefix("/expsortmod", sort_order="ascend", sort_target="mod"):
        keys += remove_prefix(meta["key"], "/expsortmod/")

    assert keys == initial_keys.encode("latin-1")

    reverse_keys = b""
    for value, meta in await client.get_prefix("/expsortmod", sort_order="descend", sort_target="mod"):
        reverse_keys += remove_prefix(meta["key"], "/expsortmod/")

    assert reverse_keys == "".join(reversed(initial_keys)).encode("latin-1")
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_replace_success():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    key = "/doot/thing" + str(uuid.uuid4())
    await client.put(key, "toot")
    status = await client.replace(key, "toot", "doot")
    v = await client.get(key)
    assert [b"doot"] == v
    assert status
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_replace_fail():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    key = "/doot/thing" + str(uuid.uuid4())
    await client.put(key, "boot")
    status = await client.replace(key, "toot", "doot")
    v = await client.get(key)
    assert [b"boot"] == v
    assert not status
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_lease():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    lease = await client.lease(ttl=60)
    assert lease is not None

    ttl = await lease.ttl()
    assert 0 <= ttl <= 60

    keys = await lease.keys()
    assert [] == keys

    ttl = await lease.refresh()
    assert 0 <= ttl <= 60

    assert await lease.revoke()
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_lease_with_keys():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    lease = await client.lease(ttl=60)
    assert lease is not None

    assert await client.put("foo12", "bar12", lease)
    assert await client.put("foo13", "bar13", lease)

    keys = await lease.keys()
    assert 2 == len(keys)
    assert b"foo12" in keys
    assert b"foo13" in keys

    assert [b"bar12"] == await client.get("foo12")
    assert [b"bar13"] == await client.get("foo13")

    assert await lease.revoke()
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_watch_key():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    key = "/%s-watch_key/watch" % str(uuid.uuid4())

    async def update_etcd(v):
        await client.put(key, v)
        out = await client.get(key)
        assert [v.encode("latin-1")] == out

    async def update_key():
        # sleep to make watch can get the event
        await asyncio.sleep(0.5)
        await update_etcd("0")
        await asyncio.sleep(0.1)
        await update_etcd("1")
        await asyncio.sleep(0.1)
        await update_etcd("2")
        await asyncio.sleep(0.1)
        await update_etcd("3")
        await asyncio.sleep(0.1)

    t = asyncio.create_task(update_key())

    change_count = 0
    events_iterator, cancel = await client.watch(key)
    async for event in events_iterator:
        assert event["kv"]["key"] == key.encode("latin-1")
        assert event["kv"]["value"] == str(change_count).encode("latin-1")

        # if cancel worked, we should not receive event 3
        assert event["kv"]["value"] != b"3"

        change_count += 1
        if change_count > 2:
            # if cancel not work, we will block in this for-loop forever
            await cancel()

    await t
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_watch_prefix():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    key = "/%s-watch_prefix/watch/prefix/" % str(uuid.uuid4())

    async def update_etcd(v):
        await client.put(key + v, v)
        out = await client.get(key + v)
        assert [v.encode("latin-1")] == out

    async def update_key():
        # sleep to make watch can get the event
        await asyncio.sleep(0.5)
        await update_etcd("0")
        await asyncio.sleep(0.1)
        await update_etcd("1")
        await asyncio.sleep(0.1)
        await update_etcd("2")
        await asyncio.sleep(0.1)
        await update_etcd("3")
        await asyncio.sleep(0.1)

    t = asyncio.create_task(update_key())

    change_count = 0
    events_iterator, cancel = await client.watch_prefix(key)
    async for event in events_iterator:
        if not event["kv"]["key"].startswith(key.encode("latin-1")):
            continue
        assert event["kv"]["key"] == ("%s%s" % (key, change_count)).encode("latin-1")
        assert event["kv"]["value"] == str(change_count).encode("latin-1")

        # if cancel worked, we should not receive event 3
        assert event["kv"]["value"] != b"3"

        change_count += 1
        if change_count > 2:
            # if cancel not work, we will block in this for-loop forever
            await cancel()

    await t


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_sequential_watch_prefix_once():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    try:
        await client.watch_prefix_once("/doot/", 1)
    except exceptions.WatchTimedOut:
        pass
    try:
        await client.watch_prefix_once("/doot/", 1)
    except exceptions.WatchTimedOut:
        pass
    try:
        await client.watch_prefix_once("/doot/", 1)
    except exceptions.WatchTimedOut:
        pass
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_lock_acquire_release():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    async with await client.lock(ttl=60) as lock:
        ttl = await lock.refresh()
        assert 0 <= ttl <= 60
    assert not await lock.is_acquired()

    async with await client.lock(ttl=60) as lock:
        assert not await lock.acquire()
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_locks():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    lock = await client.lock(id="xyz-%s" % perf_counter(), ttl=60)
    assert lock is not None

    assert await lock.acquire()
    assert lock.uuid is not None

    ttl = await lock.refresh()
    assert 0 <= ttl <= 60

    assert await lock.is_acquired()
    assert await lock.release()
    assert not await lock.release()
    assert not await lock.is_acquired()
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_create_success():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    key = "/foo/unique" + str(uuid.uuid4())
    # Verify that key is empty
    assert [] == await client.get(key)

    status = await client.create(key, "bar")
    # Verify that key is 'bar'
    assert [b"bar"] == await client.get(key)
    assert status
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_create_fail():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    key = "/foo/" + str(uuid.uuid4())
    # Assign value to the key
    await client.put(key, "bar")
    assert [b"bar"] == await client.get(key)

    status = await client.create(key, "goo")
    # Verify that key is still 'bar'
    assert [b"bar"] == await client.get(key)
    assert not status
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_create_with_lease_success():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    key = "/foo/unique" + str(uuid.uuid4())
    # Verify that key is empty
    assert [] == await client.get(key)
    lease = await client.lease()

    status = await client.create(key, "bar", lease=lease)
    # Verify that key is 'bar'
    assert [b"bar"] == await client.get(key)
    assert status
    keys = await lease.keys()
    assert 1 == len(keys)
    assert key.encode("latin-1") in keys
    await client.close()


async def _post_key(client, key_name, provide_value=True):
    payload = {"key": utils._encode(key_name)}
    if provide_value:
        payload["value"] = utils._encode(key_name)
    await client.post(client.get_url("/kv/put"), json=payload)


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_keys_with_metadata_and_value():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    test_key_value = b"some_key"
    await _post_key(client, test_key_value)
    result = await client.get(test_key_value, metadata=True)
    assert len(result) > 0
    value, metadata = result[0]
    assert value == test_key_value
    await client.close()


@pytest.mark.etcd
@pytest.mark.asyncio
async def test_client_keys_with_metadata_and_no_value():
    client = AsyncEtcd3Client(host=ETCD_HOST, api_path=API_PATH)
    value_is_not_set_default = b""
    test_key = b"some_key"
    await _post_key(client, test_key, provide_value=False)
    result = await client.get(test_key, metadata=True)
    assert len(result) > 0
    value, metadata = result[0]
    assert value == value_is_not_set_default
    await client.close()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pytest

from async_etcd3gw import AsyncEtcd3Client


@pytest.mark.asyncio
async def test_client_default():
    client = AsyncEtcd3Client()
    assert "http://localhost:2379%slease/grant" % client.api_path == client.get_url("/lease/grant")
    await client.close()


@pytest.mark.asyncio
async def test_client_ipv4():
    client = AsyncEtcd3Client(host="127.0.0.1")
    assert "http://127.0.0.1:2379%slease/grant" % client.api_path == client.get_url("/lease/grant")
    await client.close()


@pytest.mark.asyncio
async def test_client_ipv6():
    client = AsyncEtcd3Client(host="::1")
    assert "http://[::1]:2379%slease/grant" % client.api_path == client.get_url("/lease/grant")
    await client.close()


@pytest.mark.asyncio
async def test_client_api_path():
    client = AsyncEtcd3Client(host="127.0.0.1", api_path="/v3/")
    assert "http://127.0.0.1:2379/v3/lease/grant" == client.get_url("/lease/grant")
    await client.close()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#    Derived from:
#    https://opendev.org/openstack/etcd3gw/src/tag/2.1.0/etcd3gw/lock.py

import uuid

from .utils import DEFAULT_TIMEOUT, LOCK_PREFIX, _encode

__all__ = ["AsyncLock"]


class AsyncLock(object):
    def __init__(self, name, ttl=DEFAULT_TIMEOUT, async_client=None):
        """Create a lock using the given name with specified timeout

        This class represents a lock that can be acquired or released on a key in the key-value store.
        A lock has a name and a time to live (TTL) that specifies the expiration time of the lock.

        Args:
            name (str): The name of the lock.
            ttl (int): The timeout for the lock in seconds.
            async_client (AsyncClient): The async client object that communicates with the etcd v3 API.
        """
        self.name = name
        self.ttl = ttl
        self.async_client = async_client
        self.key = LOCK_PREFIX + self.name
        self.lease = None
        self._uuid = str(uuid.uuid1())

    @property
    def uuid(self):
        """The unique id of the lock"""
        return self._uuid

    async def acquire(self):
        """Acquire the lock."""
        self.lease = await self.async_client.lease(self.ttl)

        base64_key = _encode(self.key)
        base64_value = _encode(self._uuid)
        txn = {
            "compare": [
                {
                    "key": base64_key,
                    "result": "EQUAL",
                    "target": "CREATE",
                    "create_revision": 0,
                }
            ],
            "success": [
                {
                    "request_put": {
                        "key": base64_key,
                        "value": base64_value,
                        "lease": self.lease.id,
                    }
                }
            ],
            "failure": [{"request_range": {"key": base64_key}}],
        }
        result = await self.async_client.transaction(txn)
        if "succeeded" in result:
            return result["succeeded"]
        return False

    async def release(self):
        """Release the lock"""
        base64_key = _encode(self.key)
        base64_value = _encode(self._uuid)

        txn = {
            "compare": [
                {
                    "key": base64_key,
                    "result": "EQUAL",
                    "target": "VALUE",
                    "value": base64_value,
                }
            ],
            "success": [{"request_delete_range": {"key": base64_key}}],
        }

        result = await self.async_client.transaction(txn)
        if "succeeded" in result:
            return result["succeeded"]
        return False

    async def refresh(self):
        """Refresh the lease on the lock"""
        return await self.lease.refresh()

    async def is_acquired(self):
        """Check if the lock is acquired"""
        values = await self.async_client.get(self.key)
        return self._uuid.encode("latin-1") in values

    async def __aenter__(self):
        """Use the lock as a contextmanager"""
        await self.acquire()
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        await self.release()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#    Derived from:
#    https://opendev.org/openstack/etcd3gw/src/tag/2.1.0/etcd3gw/client.py

import asyncio
import json
import ssl
import threading
import uuid

import aiohttp

from .async_lease import AsyncLease
from .async_lock import AsyncLock
from .async_watch import AsyncWatcher
from .exceptions import (
    ConnectionFailedError,
    ConnectionTimeoutError,
    WatchTimedOut,
    get_exception,
)
from .utils import DEFAULT_TIMEOUT, _decode, _encode, _increment_last_byte

__all__ = [
    "AsyncEtcd3Client",
    "async_client",
    "DEFAULT_API_PATH",
    "DEFAULT_PORT",
    "DEFAULT_PROTOCOL",
]

_SORT_ORDER = ["none", "ascend", "descend"]
_SORT_TARGET = ["key", "version", "create", "mod", "value"]
DEFAULT_API_PATH = "/v3/"
DEFAULT_PORT = 2379
DEFAULT_PROTOCOL = "http"

# gRPC gateway endpoints:
#
# - etcd v3.2 or before uses only [CLIENT-URL]/v3alpha/*
# - etcd v3.3 uses [CLIENT-URL]/v3beta/* while keeping [CLIENT-URL]/v3alpha/*
# - etcd v3.4 uses [CLIENT-URL]/v3/* while keeping [CLIENT-URL]/v3beta/*
#                  [CLIENT-URL]/v3alpha/* is deprecated
# - etcd v3.5 or later uses only [CLIENT-URL]/v3/*
#                  [CLIENT-URL]/v3beta/* is deprecated


class AsyncEtcd3Client(object):
    def __init__(
        self,
        host="localhost",
        port=DEFAULT_PORT,
        protocol=DEFAULT_PROTOCOL,
        ca_cert=None,
        cert_key=None,
        cert_cert=None,
        timeout=None,
        api_path=DEFAULT_API_PATH,
    ):
        """Construct an client to talk to etcd3's grpc-gateway's /v3 HTTP API

        Args:
            host (str): The etcd hostname.
            port (int): The etcd port.
            protocol (str): The etcd protocol (http/https).
            timeout (int): The default timeout.
            api_path (str): The api path.
        """
        self.host = host
        self.port = port
        self.protocol = protocol
        self.api_path = api_path

        if timeout is not None:
            self.timeout = aiohttp.ClientTimeout(total=None, sock_connect=timeout, sock_read=timeout)
        else:
            self.timeout = None

        if ca_cert is not None:
            ssl_ctx = ssl.create_default_context(cafile=ca_cert)
            if cert_cert is not None and cert_key is not None:
                ssl_ctx.load_cert_chain(cert_cert, cert_key)
            self.connector = aiohttp.TCPConnector(ssl_context=ssl_ctx)
        elif cert_cert is not None and cert_key is not None:
            ssl_ctx = ssl.create_default_context()
            ssl_ctx.load_cert_chain(cert_cert, cert_key)
            self.connector = aiohttp.TCPConnector(ssl_context=ssl_ctx)
        else:
            self.connector = None

        self.client_session_kwargs = {}
        if self.timeout is not None:
            self.client_session_kwargs["timeout"] = self.timeout
        if self.connector is not None:
            self.client_session_kwargs["connector"] = self.connector

        self.session = aiohttp.ClientSession(**self.client_session_kwargs)

    def get_url(self, path):
        """Construct a full url to the v3 API given a specific path

        This method takes a path as an argument and returns a url that points to the v3 API of the host.

        Args:
            path (str): The path to append to the base url.

        Returns:
            str: The full url to the v3 API with the given path.
        """
        host = f"[{ self.host }]" if ":" in self.host else self.host
        base_url = f"{ self.protocol }://{ host }:{ self.port }"
        return base_url + self.api_path + path.lstrip("/")

    async def post(self, *args, **kwargs):
        """Helper method for HTTP POST

        This method performs an asynchronous HTTP POST request using aiohttp.
        It takes arbitrary arguments and keyword arguments that are passed to the session.post method.
        It returns a json response if the request is successful, or raises an exception if the request
        fails or times out.

        Example:
            >>> self.client_session_kwargs = {"timeout": 10}
            >>> await self.post("https://example.com/api", data={"key": "value"})
            {'result': 'ok'}

        Args:
            *args: Positional arguments for the session.post method.
            **kwargs: Keyword arguments for the session.post method.

        Returns:
            dict: The json response from the server.

        Raises:
            AsyncEtcd3Exception: If the status code is not 200 and not in the predefined exceptions.
            ConnectionTimeoutError: If the request times out.
            ConnectionFailedError: If the connection fails.
        """
        try:
            async with self.session.post(*args, **kwargs) as resp:
                ex = get_exception(resp.status, resp.text, resp.reason)
                if ex is not None:
                    raise ex
                return await resp.json()
        except asyncio.TimeoutError as ex:
            raise ConnectionTimeoutError(str(ex))
        except aiohttp.ClientConnectionError as ex:
            raise ConnectionFailedError(str(ex))

    async def status(self):
        """Status gets the status of the etcd cluster member.

        Returns:
            dict: The json response from the server.
        """
        return await self.post(self.get_url("/maintenance/status"), json={})

    async def members(self):
        """Lists all the members in the cluster.

        Returns:
            dict: The json response from the server.
        """
        result = await self.post(self.get_url("/cluster/member/list"), json={})
        return result["members"]

    async def lease(self, ttl=DEFAULT_TIMEOUT):
        """Create a AsyncLease object given a timeout

        This method performs an post request to the etcd v3 API to create a lease
        with a specified time to live (TTL). It takes a TTL as an argument,
        and returns an AsyncLease object that represents the lease.
        The lease can be used to attach keys to it, keep it alive, or revoke it.

        Example:
            >>> lease = await self.lease(ttl=10)
            >>> await self.put("/foo", "bar", lease=lease)
            True

        Args:
            ttl (int): The timeout for the lease in seconds.

        Returns:
            AsyncLease: The lease object.

        """
        result = await self.post(self.get_url("/lease/grant"), json={"TTL": ttl, "ID": 0})
        return AsyncLease(int(result["ID"]), async_client=self)

    async def lock(self, id=None, ttl=DEFAULT_TIMEOUT):
        """Create a Lock object given an ID and timeout

        This method creates a lock with a specified ID and TTL.
        It takes an ID and a TTL as arguments, and returns an AsyncLock object that represents the lock.
        The lock can be used to acquire or release the lock on a key.

        Example:
            >>> lock = await self.lock(id="mylock", ttl=10)
            >>> await lock.acquire("foo")
            True

        Args:
            id (str): The ID for the lock. Creates a new UUID if not provided.
            ttl (int): The timeout for the lock in seconds.

        Returns:
            AsyncLock: The lock object.
        """
        if id is None:
            id = str(uuid.uuid4())
        return AsyncLock(id, ttl=ttl, async_client=self)

    async def create(self, key, value, lease=None):
        """Atomically create the given key only if the key doesn't exist.

        This verifies that the create_revision of a key equales to 0, then
        creates the key with the value.
        This operation takes place in a transaction.

        Example:
            >>> await self.create("/foo", "bar")
            True
            >>> await self.create("/foo", "baz")
            False

        Args:
            key (str): The key to create.
            value (str): The value to store.
            lease (Lease): The lease object that specifies the expiration time of the key.

        Returns:
            bool: True if the create was successful, False otherwise.
        """
        base64_key = _encode(key)
        base64_value = _encode(value)
        txn = {
            "compare": [
                {
                    "key": base64_key,
                    "result": "EQUAL",
                    "target": "CREATE",
                    "create_revision": 0,
                }
            ],
            "success": [
                {
                    "request_put": {
                        "key": base64_key,
                        "value": base64_value,
                    }
                }
            ],
            "failure": [],
        }
        if lease:
            txn["success"][0]["request_put"]["lease"] = lease.id
        result = await self.transaction(txn)
        if "succeeded" in result:
            return result["succeeded"]
        return False

    async def put(self, key, value, lease=None, metadata=False):
        """Put puts the given key into the key-value store.

        This method performs an asynchronous HTTP POST request to the etcd v3 API to store
        a key-value pair. It takes a key and a value as arguments, and optionally a lease object
        that specifies the expiration time of the key.
        A put request increments the revision of the key-value store
        and generates one event in the event history.

        Example:
            >>> await self.put("/foo", "bar")
            True

        Args:
            key (str): The key to store.
            value (str): The value to store.
            lease (Lease): The lease object that specifies the expiration time of the key.
            metadata (bool): Whether to return the metadata.

        Returns:
            bool, dict: If metadata is True, the json response from the server else True if the put was successful.
        """
        payload = {"key": _encode(key), "value": _encode(value)}
        if lease:
            payload["lease"] = lease.id
        result = await self.post(self.get_url("/kv/put"), json=payload)
        if metadata:
            return result
        else:
            return True

    async def get(self, key, metadata=False, sort_order=None, sort_target=None, **kwargs):
        """Range gets the keys in the range from the key-value store.

         This method performs an asynchronous HTTP POST request to the etcd v3 API to retrieve
         a range of keys from the key-value store.

        Example:
            >>> await self.get("/foo")
            [b'bar']
            >>> await self.get("/foo", metadata=True)
            [(b'bar', {'key': 'foo', 'create_revision': 1, 'mod_revision': 1, 'version': 1})]
            >>> await self.get("/foo", sort_order="descend", sort_target="version")
            [b'baz', b'bar']

        Args:
            key (str): The key to query.
            metadata (bool): Whether to return the metadata of the keys.
            sort_order (str): The order of the keys. One of "ascend" or "descend" or None.
            sort_target (str): The criteria for sorting. One of "key", "version", "create", "mod" or "value".
            **kwargs: Other keyword arguments for the request payload.

        Returns:
            list: A list of values or a list of tuples of value and metadata.
        """
        try:
            order = 0
            if sort_order:
                order = _SORT_ORDER.index(sort_order)
        except ValueError:
            raise ValueError('sort_order must be one of "ascend" or "descend"')

        try:
            target = 0
            if sort_target:
                target = _SORT_TARGET.index(sort_target)
        except ValueError:
            raise ValueError('sort_target must be one of "key", ' '"version", "create", "mod" or "value"')

        payload = {
            "key": _encode(key),
            "sort_order": order,
            "sort_target": target,
        }
        payload.update(kwargs)
        result = await self.post(self.get_url("/kv/range"), json=payload)
        if "kvs" not in result:
            return []

        if metadata:

            def value_with_metadata(item):
                item["key"] = _decode(item["key"])
                value = _decode(item.pop("value", ""))
                return value, item

            return [value_with_metadata(item) for item in result["kvs"]]

        return [_decode(item.get("value", "")) for item in result["kvs"]]

    async def get_all(self, sort_order=None, sort_target="key"):
        """Get all keys currently stored in etcd.

        Returns:
            list: A sequence of (value, metadata) tuples
        """
        return await self.get(
            key=_encode(b"\0"),
            metadata=True,
            sort_order=sort_order,
            sort_target=sort_target,
            range_end=_encode(b"\0"),
        )

    async def get_prefix(self, key_prefix, sort_order=None, sort_target=None):
        """Get a range of keys with a prefix.

        Args:
            key_prefix (str): The prefix to query.
            sort_order (str): The order of the keys. One of "ascend" or "descend" or None.
            sort_target (str): The criteria for sorting. One of "key", "version", "create", "mod" or "value".

        Returns:
            list: A sequence of (value, metadata) tuples
        """
        return await self.get(
            key_prefix,
            metadata=True,
            range_end=_encode(_increment_last_byte(key_prefix)),
            sort_order=sort_order,
            sort_target=sort_target,
        )

    async def replace(self, key, initial_value, new_value, metadata=False):
        """Atomically replace the value of a key with a new value.

        This compares the current value of a key, then replaces it with a new
        value if it is equal to a specified value. This operation takes place
        in a transaction.

        Example:
            >>> await self.replace("/foo", "bar", "baz")
            True
            >>> await self.replace("/foo", "bar", "qux")
            False

        Args:
            key (str): The key to replace.
            initial_value (str): The old value to compare.
            new_value (str): The new value to swap.
            metadata (bool): Whether to return the metadata.

        Returns:
            bool, dict: If metadata is True, the json response from the server else True if the replace was successful.
        """
        base64_key = _encode(key)
        base64_initial_value = _encode(initial_value)
        base64_new_value = _encode(new_value)
        txn = {
            "compare": [
                {
                    "key": base64_key,
                    "result": "EQUAL",
                    "target": "VALUE",
                    "value": base64_initial_value,
                }
            ],
            "success": [
                {
                    "request_put": {
                        "key": base64_key,
                        "value": base64_new_value,
                    }
                }
            ],
            "failure": [],
        }
        result = await self.transaction(txn)
        if metadata:
            return result
        elif "succeeded" in result:
            return result["succeeded"]
        return False

    async def delete(self, key, metadata=False, **kwargs):
        """DeleteRange deletes the given range from the key-value store.

        A delete request increments the revision of the key-value store and
        generates a delete event in the event history for every deleted key.

        Args:
            key (str): The key to delete.
            metadata (bool): Whether to return the metadata.

        Returns:
            bool, dict: If metadata is True, the json response from the server else True if the delete was successful.
        """
        payload = {
            "key": _encode(key),
        }
        payload.update(kwargs)

        result = await self.post(self.get_url("/kv/deleterange"), json=payload)
        if metadata:
            return result
        else:
            return "deleted" in result

    async def delete_prefix(self, key_prefix, metadata=False):
        """Delete a range of keys with a prefix in etcd."""
        return await self.delete(key_prefix, range_end=_encode(_increment_last_byte(key_prefix)), metadata=metadata)

    async def transaction(self, txn):
        """Txn processes multiple requests in a single transaction.

        A txn request increments the revision of the key-value store and
        generates events with the same revision for every completed request.
        It is not allowed to modify the same key several times within one txn.

        Args:
            txn: The json request payload.
        Returns:
            dict: The json response from the server.
        """
        return await self.post(self.get_url("/kv/txn"), data=json.dumps(txn))

    async def watch(self, key, **kwargs):
        """Watch a key.

        This method creates a watch request to the etcd v3 API to monitor a key for changes.
        It returns a tuple of events_iterator and cancel.
        Use events_iterator to get the events of key changes and cancel to cancel the watch request.

        Args:
            key (str): The key to watch.
            **kwargs: Other keyword arguments for the watch request.

        Returns:
            tuple: A tuple of events_iterator and cancel.
        """
        event_queue = asyncio.Queue()

        async def callback(event):
            await event_queue.put(event)

        w = AsyncWatcher(async_client=self, key=key, callback=callback, **kwargs)
        await w.start()
        canceled = threading.Event()

        async def cancel():
            canceled.set()
            await event_queue.put(None)
            await w.stop()

        async def iterator():
            while not canceled.is_set():
                event = await event_queue.get()
                if isinstance(event, Exception):
                    canceled.set()
                    raise event
                if event is None:
                    canceled.set()
                if not canceled.is_set():
                    yield event

        return iterator(), cancel

    async def watch_prefix(self, key_prefix, **kwargs):
        """The same as ``watch``, but watches a range of keys with a prefix."""
        kwargs["range_end"] = _increment_last_byte(key_prefix)
        return await self.watch(key_prefix, **kwargs)

    async def watch_once(self, key, timeout=None, **kwargs):
        """Watch a key and stops after the first event.

        Args:
            key (str): The key to watch.
            timeout (int): The timeout in seconds. Optional.
            **kwargs: Other keyword arguments for the watch request.

        Returns:
            dict: The first event of the key change.

        Raises:
            WatchTimedOut: If the timeout is reached.
        """
        event_queue = asyncio.Queue()

        async def callback(event):
            await event_queue.put(event)

        w = AsyncWatcher(async_client=self, key=key, callback=callback, **kwargs)
        await w.start()
        try:
            return await asyncio.wait_for(event_queue.get(), timeout=timeout)
        except asyncio.exceptions.TimeoutError:
            raise WatchTimedOut()
        finally:
            await w.stop()

    async def watch_prefix_once(self, key_prefix, timeout=None, **kwargs):
        """Watches a range of keys with a prefix, similar to watch_once"""
        kwargs["range_end"] = _increment_last_byte(key_prefix)
        return await self.watch_once(key_prefix, timeout=timeout, **kwargs)

    async def close(self):
        """Close the underlying connector and release all acquired resources."""
        return await self.session.close()

    async def __aenter__(self):
        """Use the AsyncEtcd3Client as a contextmanager"""
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        """Use the AsyncEtcd3Client as a contextmanager"""
        await self.close()


def async_client(
    host="localhost",
    port=2379,
    ca_cert=None,
    cert_key=None,
    cert_cert=None,
    timeout=None,
    protocol="http",
    api_path=DEFAULT_API_PATH,
):
    """Return an instance of AsyncEtcd3Client"""
    return AsyncEtcd3Client(
        host=host,
        port=port,
        ca_cert=ca_cert,
        cert_key=cert_key,
        cert_cert=cert_cert,
        timeout=timeout,
        api_path=api_path,
        protocol=protocol,
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#    Derived from:
#    https://opendev.org/openstack/etcd3gw/src/tag/2.1.0/etcd3gw/lease.py

from .utils import _decode

__all__ = ["AsyncLease"]


class AsyncLease(object):
    def __init__(self, id, async_client=None):
        """Lease object for expiring keys.

        This class represents a lease that can be attached to keys in the key-value store.
        A lease has an ID and a time to live (TTL) that specifies the expiration time of the keys.
        A lease can be revoked to delete all the attached keys, or queried to get its information.

        Args:
            id (int): The ID of the lease.
            async_client (AsyncClient): The AsyncClient object that communicates with the etcd v3 API.
        """
        self.id = id
        self.async_client = async_client

    async def revoke(self):
        """Revoke a lease.

        This method performs an asynchronous HTTP POST request to the etcd v3 API to revoke a lease.
        All keys attached to the lease will expire and be deleted.
        It returns True if the request is successful.

        Example:
            >>> await lease.revoke()
            True

        Returns:
            bool: True if the request is successful.
        """
        return True

    async def ttl(self):
        """Retrieve lease information.

        This method performs an asynchronous HTTP POST request to the etcd v3 API to retrieve
        lease information. It returns the TTL of the lease in seconds.

        Example:
            >>> await lease.ttl()
            60

        Returns:
            int: The TTL of the lease in seconds.
        """
        result = await self.async_client.post(self.async_client.get_url("/kv/lease/timetolive"), json={"ID": self.id})
        return int(result["TTL"])

    async def refresh(self):
        """Keep the lease alive.

        This method performs an asynchronous HTTP POST request to the etcd v3 API to keep the lease alive.
        By streaming keep alive requests from the client to the server and streaming keep alive responses
        from the server to the client. It returns the new TTL for the lease in seconds.
        If the lease was already expired, then the TTL field is absent in the response and the function
        returns -1 according to etcd documentation.

        Example:
            >>> await lease.refresh()
            10

        Returns:
            int: The new TTL for the lease in seconds, or -1 if the lease was expired.
        """
        result = await self.async_client.post(self.async_client.get_url("/lease/keepalive"), json={"ID": self.id})
        return int(result["result"].get("TTL", -1))

    async def keys(self):
        """Get the keys associated with this lease.

        This method performs an asynchronous HTTP POST request to the etcd v3 API to get the keys
        associated with this lease. It returns a list of keys that are attached to the lease.

        Example:
            >>> await lease.keys()
            [b'/foo', b'/bar']

        Returns:
            list: A list of keys that are attached to the lease.
        """
        result = await self.async_client.post(
            self.async_client.get_url("/kv/lease/timetolive"),
            json={"ID": self.id, "keys": True},
        )
        keys = result["keys"] if "keys" in result else []
        return [_decode(key) for key in keys]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#    Derived from:
#    https://opendev.org/openstack/etcd3gw/src/tag/2.1.0/etcd3gw/exceptions.py

__all__ = [
    "get_exception",
    "AsyncEtcd3Exception",
    "WatchTimedOut",
    "InternalServerError",
    "ConnectionFailedError",
    "ConnectionTimeoutError",
    "PreconditionFailedError",
    "_EXCEPTIONS_BY_CODE",
]


class AsyncEtcd3Exception(Exception):
    def __init__(self, detail_text=None, *args):
        super().__init__(*args)
        self.detail_text = detail_text


class WatchTimedOut(AsyncEtcd3Exception):
    pass


class InternalServerError(AsyncEtcd3Exception):
    pass


class ConnectionFailedError(AsyncEtcd3Exception):
    pass


class ConnectionTimeoutError(AsyncEtcd3Exception):
    pass


class PreconditionFailedError(AsyncEtcd3Exception):
    pass


_EXCEPTIONS_BY_CODE = {
    500: InternalServerError,
    503: ConnectionFailedError,
    408: ConnectionTimeoutError,
    504: ConnectionTimeoutError,
    412: PreconditionFailedError,
}


def get_exception(status, text, reason):
    if status in _EXCEPTIONS_BY_CODE:
        return _EXCEPTIONS_BY_CODE[status](text, reason)
    if status != 200:
        return AsyncEtcd3Exception(text, reason)
    return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#    Derived from:
#    https://opendev.org/openstack/etcd3gw/src/tag/2.1.0/etcd3gw/watch.py

import asyncio
import json

import aiohttp

from .exceptions import get_exception
from .utils import _decode, _encode

__all__ = ["AsyncWatcher"]


class AsyncWatcher(object):
    KW_ARGS = ["start_revision", "progress_notify", "filters", "prev_kv"]
    KW_ENCODED_ARGS = ["range_end"]

    def __init__(self, async_client, key, callback, **kwargs):
        self.async_client = async_client
        self.callback = callback
        create_watch = {"key": _encode(key)}
        for arg in kwargs:
            if arg in self.KW_ARGS:
                create_watch[arg] = kwargs[arg]
            elif arg in self.KW_ENCODED_ARGS:
                create_watch[arg] = _encode(kwargs[arg])
        self.create_request = {"create_request": create_watch}

    async def start(self):
        "Start the watch task"
        self.event = asyncio.Event()
        self.exception = None
        self.watch_task = asyncio.create_task(self.watch())
        await self.event.wait()
        if self.exception is not None:
            raise self.exception

    async def stop(self):
        "Request the watch task to be cancelled"
        self.watch_task.cancel()
        try:
            await self.watch_task
        except asyncio.CancelledError:
            pass

    async def process_chunk(self, line):
        "Process a chuck"
        decoded_line = line.decode("utf-8")
        try:
            payload = json.loads(decoded_line)
        except json.JSONDecodeError as ex:
            raise ex
        if "error" in payload:
            raise get_exception(
                payload["error"].get("http_code", 500), payload["error"].get("http_status", ""), payload["error"].get("message", "")
            )
        if "created" in payload["result"]:
            if not payload["result"]["created"]:
                raise Exception("Unable to create watch")
        if "events" in payload["result"]:
            for event in payload["result"]["events"]:
                event["kv"]["key"] = _decode(event["kv"]["key"])
                if "value" in event["kv"]:
                    event["kv"]["value"] = _decode(event["kv"]["value"])
                await self.callback(event)

    async def read_and_process_chunk(self, content):
        while True:
            if content._exception is not None:
                raise content._exception

            if not content._buffer and not content._eof:
                await content._wait("readany")

            (data, end_of_http_chunk) = await content.readchunk()
            while not end_of_http_chunk:
                (more_data, end_of_http_chunk) = await content.readchunk()
                data += more_data

            await self.process_chunk(data)

    async def watch(self):
        try:
            timeout = aiohttp.ClientTimeout(total=None, sock_read=None)
            async with self.async_client.session.post(
                self.async_client.get_url("/watch"), json=self.create_request, timeout=timeout
            ) as resp:
                ex = get_exception(resp.status, resp.text, resp.reason)
                if ex:
                    raise ex
                self.event.set()
                await self.read_and_process_chunk(resp.content)
        except Exception as ex:
            self.exception = ex
            self.event.set()
            await self.callback(ex)
